#!/usr/bin/env python3
"""
Created on Sun Apr 25 13:32:04 2021

@author: vkapoor
"""
import csv
import glob
import os
import time

# from IPython.display import clear_output
from pathlib import Path

import numpy as np
from keras.models import load_model
from natsort import natsorted
from tifffile import imread, imwrite
from tqdm import tqdm

from oneat.NEATModels.neat_lstm import NEATLRNet
from oneat.NEATModels.nets import Concat
from oneat.NEATUtils.utils import (
    microscope_dynamic_nms,
    normalizeFloatZeroOne,
    yoloprediction,
)


class NEATPredict(NEATLRNet):
    def __init__(self, config, model_dir, model_name, catconfig, cordconfig):

        super().__init__(
            config=config,
            model_dir=model_dir,
            model_name=model_name,
            catconfig=catconfig,
            cordconfig=cordconfig,
        )

    def predict_microscope(
        self,
        imagedir,
        Z_imagedir,
        start=0,
        Z_start=0,
        roi_start=0,
        roi_end=1,
        movie_name_list={},
        movie_input={},
        Z_movie_name_list=[],
        Z_movie_input=[],
        fileextension="*TIF",
        nb_prediction=3,
        n_tiles=(1, 1),
        Z_n_tiles=(1, 2, 2),
        overlap_percent=0.6,
        event_threshold=0.5,
        event_confidence=0.5,
        iou_threshold=0.01,
        projection_model=None,
        delay_projection=4,
        fidelity=4,
        jumpindex=1,
        normalize=True,
        optional_name=None,
        center_oneat=True,
        nms_function="iou",
    ):

        self.imagedir = imagedir
        self.basedirResults = self.imagedir + "/" + "live_results"
        Path(self.basedirResults).mkdir(exist_ok=True)
        # Recurrsion variables
        self.movie_name_list = movie_name_list
        self.movie_input = movie_input
        self.Z_movie_name_list = Z_movie_name_list
        self.delay_projection = delay_projection
        self.Z_movie_input = Z_movie_input
        self.Z_imagedir = Z_imagedir
        self.start = start
        self.jumpindex = jumpindex
        self.fidelity = fidelity
        self.optional_name = optional_name
        self.Z_start = Z_start
        self.projection_model = projection_model
        self.nb_prediction = nb_prediction
        self.fileextension = fileextension
        self.n_tiles = n_tiles
        self.nms_function = nms_function
        self.roi_start = roi_start
        self.roi_end = roi_end
        self.Z_n_tiles = Z_n_tiles
        self.overlap_percent = overlap_percent
        self.iou_threshold = iou_threshold
        self.event_threshold = event_threshold
        self.event_confidence = event_confidence
        self.normalize = normalize
        self.center_oneat = center_oneat

        self.model = self._build()

        # Z slice folder listener
        while 1:

            Z_Raw_path = os.path.join(self.Z_imagedir, self.fileextension)
            Z_filesRaw = glob.glob(Z_Raw_path)
            Z_filesRaw = natsorted(Z_filesRaw)

            Raw_path = os.path.join(self.imagedir, "*tif")
            filesRaw = glob.glob(Raw_path)
            filesRaw = natsorted(filesRaw)

            for Z_movie_name in Z_filesRaw:
                Z_Name = os.path.basename(os.path.splitext(Z_movie_name)[0])
                # Check for unique filename
                if self.optional_name is not None:
                    if (
                        Z_Name not in self.Z_movie_name_list
                        and self.optional_name in Z_Name
                    ):
                        self.Z_movie_name_list.append(Z_Name)
                        self.Z_movie_input.append(Z_movie_name)
                else:
                    if Z_Name not in self.Z_movie_name_list:
                        self.Z_movie_name_list.append(Z_Name)
                        self.Z_movie_input.append(Z_movie_name)

            for movie_name in filesRaw:
                Name = os.path.basename(os.path.splitext(movie_name)[0])
                # Check for unique filename
                if Name not in self.movie_name_list:
                    self.movie_name_list[Name] = Name
                    self.movie_input[Name] = movie_name

                    total_movies = len(self.movie_input)

            doproject = True

            for i in range(len(self.Z_movie_name_list)):

                Z_Name = self.Z_movie_name_list[i]
                Z_path = self.Z_movie_input[i]

                if Z_Name in self.movie_name_list:

                    Name = self.movie_name_list[Z_Name]

                    doproject = False
                else:
                    doproject = True

                if doproject:
                    time.sleep(self.delay_projection)
                    try:
                        start_time = time.time()
                        print("Reading Z stack for projection")
                        Z_image = imread(Z_path)
                        print("Read properly")
                    except FileNotFoundError:
                        Z_image = None

                    if Z_image is not None:
                        if self.projection_model is not None:
                            print("Projecting using the projection model")
                            projection = self.projection_model.predict(
                                Z_image, "ZYX", n_tiles=Z_n_tiles
                            )
                        else:
                            print("Doing max projection")
                            projection = np.amax(Z_image, axis=0)
                        imwrite(
                            self.imagedir + "/" + Z_Name + ".tif",
                            projection.astype("float32"),
                        )
                        print(
                            f"____ Projection took {(time.time() - start_time)} seconds ____ "
                        )

                    else:
                        if Z_Name in self.Z_movie_name_list:
                            self.Z_movie_name_list.remove(Z_Name)
                        if Z_movie_name in self.Z_movie_input:
                            self.Z_movie_input.remove(Z_movie_name)

            self.movie_input_list = []
            for (k, v) in self.movie_input.items():
                self.movie_input_list.append(v)
            total_movies = len(self.movie_input_list)
            if total_movies > self.size_tminus + self.start:
                current_movies = imread(
                    self.movie_input_list[
                        self.start : self.start + self.size_tminus + 1
                    ]
                )

                current_movies_down = current_movies
                # print(current_movies_down.shape)
                print(
                    "Predicting on Movies:",
                    self.movie_input_list[
                        self.start : self.start + self.size_tminus + 1
                    ],
                )
                inputtime = self.start + self.size_tminus

                eventboxes = []
                classedboxes = {}
                smallimage = CreateVolume(
                    current_movies_down, self.size_tminus + 1, 0
                )
                if self.normalize:
                    smallimage = normalizeFloatZeroOne(smallimage, 1, 99.8)
                # Break image into tiles if neccessary
                self.image = smallimage
                print("Doing ONEAT prediction")
                start_time = time.time()
                predictions, allx, ally = self.predict_main(smallimage)
                print(
                    f"____ Prediction took {(time.time() - start_time)} seconds ____ "
                )

                # Iterate over tiles
                for p in tqdm(range(0, len(predictions))):

                    sum_time_prediction = predictions[p]

                    if sum_time_prediction is not None:
                        for i in range(0, sum_time_prediction.shape[0]):
                            time_prediction = sum_time_prediction[i]

                            boxprediction = yoloprediction(
                                ally[p],
                                allx[p],
                                time_prediction,
                                self.stride,
                                inputtime,
                                self.config,
                                self.key_categories,
                                self.key_cord,
                                self.nboxes,
                                "prediction",
                                "dynamic",
                                center_oneat=self.center_oneat,
                            )

                            if boxprediction is not None:
                                eventboxes = eventboxes + boxprediction

                for (event_name, event_label) in self.key_categories.items():

                    if event_label > 0:
                        current_event_box = []
                        for box in eventboxes:

                            event_prob = box[event_name]
                            event_confidence = box["confidence"]
                            if (
                                event_prob >= self.event_threshold
                                and event_confidence >= self.event_confidence
                            ):
                                current_event_box.append(box)
                        classedboxes[event_name] = [current_event_box]

                self.classedboxes = classedboxes
                self.eventboxes = eventboxes
                print("Performining non maximal supression")
                start_time = time.time()
                self.iou_classedboxes = classedboxes
                self.nms_microscope()
                print(
                    f"____ NMS took {(time.time() - start_time)} seconds ____ "
                )
                print("Generating ini file")
                self.to_csv_microscope()

                self.predict_microscope(
                    self.imagedir,
                    self.Z_imagedir,
                    start=self.start,
                    Z_start=self.Z_start,
                    fileextension=self.fileextension,
                    roi_start=self.roi_start,
                    roi_end=self.roi_end,
                    movie_name_list=self.movie_name_list,
                    movie_input=self.movie_input,
                    Z_movie_name_list=self.Z_movie_name_list,
                    Z_movie_input=self.Z_movie_input,
                    nb_prediction=self.nb_prediction,
                    n_tiles=self.n_tiles,
                    Z_n_tiles=self.Z_n_tiles,
                    overlap_percent=self.overlap_percent,
                    event_threshold=self.event_threshold,
                    event_confidence=self.event_confidence,
                    iou_threshold=self.iou_threshold,
                    projection_model=self.projection_model,
                    delay_projection=self.delay_projection,
                    fidelity=self.fidelity,
                    jumpindex=self.jumpindex,
                    normalize=self.normalize,
                    optional_name=self.optional_name,
                )

    def _build(self):

        model_weights = os.path.join(self.model_dir, "weights.h5")
        Model = load_model(
            model_weights,
            custom_objects={"loss": self.yolo_loss, "Concat": Concat},
        )

        return Model

    def nms_microscope(self):

        best_iou_classedboxes = {}
        self.iou_classedboxes = {}
        self.start = self.start + self.jumpindex
        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:

                best_sorted_event_box = microscope_dynamic_nms(
                    self.classedboxes,
                    event_name,
                    self.iou_threshold,
                    self.event_threshold,
                    self.imagex,
                    self.imagey,
                    self.fidelity,
                    self.nms_function,
                )

                best_iou_classedboxes[event_name] = [best_sorted_event_box]

        self.iou_classedboxes = best_iou_classedboxes

    def to_csv_microscope(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:

                xlocations = []
                ylocations = []
                scores = []
                tlocations = []
                radiuses = []
                confidences = []

                iou_current_event_boxes = self.iou_classedboxes[event_name][0]
                bbox_left_y = self.image.shape[1] * self.roi_start
                bbox_right_y = self.image.shape[1] * self.roi_end

                bbox_left_x = self.image.shape[2] * self.roi_start
                bbox_right_x = self.image.shape[2] * self.roi_end
                iou_current_event_boxes = sorted(
                    iou_current_event_boxes,
                    key=lambda x: x[event_name],
                    reverse=True,
                )

                for iou_current_event_box in iou_current_event_boxes:
                    xcenter = iou_current_event_box["xcenter"]
                    ycenter = iou_current_event_box["ycenter"]
                    tcenter = iou_current_event_box["real_time_event"]
                    score = iou_current_event_box[event_name]
                    radius = (
                        np.sqrt(
                            iou_current_event_box["height"]
                            * iou_current_event_box["height"]
                            + iou_current_event_box["width"]
                            * iou_current_event_box["width"]
                        )
                        // 2
                    )
                    confidence = iou_current_event_box["confidence"]
                    if (
                        xcenter >= bbox_left_x
                        and xcenter <= bbox_right_x
                        and ycenter >= bbox_left_y
                        and ycenter <= bbox_right_y
                    ):
                        print(round(xcenter), round(ycenter), score)
                        xlocations.append(round(xcenter))
                        ylocations.append(round(ycenter))
                        scores.append(score)
                        tlocations.append(tcenter)
                        radiuses.append(radius)
                        confidences.append(confidence)

                total_event_count = np.column_stack(
                    [
                        tlocations,
                        ylocations,
                        xlocations,
                        scores,
                        radiuses,
                        confidences,
                    ]
                )
                csvname = self.basedirResults + "/" + event_name

                ImageResults = self.basedirResults + "/" + "ImageLocations"
                Path(ImageResults).mkdir(exist_ok=True)

                event_data = []
                writer = csv.writer(open(csvname + ".csv", "a"))
                filesize = os.stat(csvname + ".csv").st_size
                if filesize < 1:
                    writer.writerow(
                        ["T", "Y", "X", "Score", "Size", "Confidence"]
                    )
                for line in total_event_count:
                    if line not in event_data:
                        event_data.append(line)
                    writer.writerows(event_data)
                    event_data = []


def CreateVolume(patch, imaget, timepoint):
    starttime = timepoint
    endtime = timepoint + imaget
    smallimg = patch[starttime:endtime, :]

    return smallimg



import csv
import glob
import os
import time

# from IPython.display import clear_output
from pathlib import Path

# from qtpy.QtCore import Qt
# from qtpy.QtWidgets import QComboBox, QPushButton, QSlider
import numpy as np
import pandas as pd
import tensorflow as tf
from keras.models import load_model
from scipy.ndimage import gaussian_filter
from scipy.optimize import curve_fit
from tifffile import imread, imwrite
from tqdm import tqdm

from oneat.NEATModels.neat_focus import NEATFocus
from oneat.NEATModels.nets import Concat
from oneat.NEATUtils.utils import (
    focyoloprediction,
    normalizeFloatZeroOne,
    simpleaveragenms,
)

# from napari.qt.threading import thread_worker
# from matplotlib.backends.backend_qt5agg import \
# FigureCanvasQTAgg as FigureCanvas
Boxname = "ImageIDBox"
EventBoxname = "EventIDBox"


class NEATFocusPredict(NEATFocus):
    def __init__(self, config, model_dir, catconfig, cordconfig):

        super().__init__(
            config=config,
            model_dir=model_dir,
            catconfig=catconfig,
            cordconfig=cordconfig,
        )

    def predict(
        self,
        imagedir: str,
        Z_imagedir: str,
        Z_movie_name_list: list,
        Z_movie_input: list,
        start: int,
        fileextension: str = "*TIF",
        nb_prediction: int = 3,
        Z_n_tiles: tuple = (1, 2, 2),
        overlap_percent: float = 0.6,
        normalize: bool = True,
    ):

        self.imagedir = imagedir
        self.basedirResults = self.imagedir + "/" + "live_results"
        Path(self.basedirResults).mkdir(exist_ok=True)
        # Recurrsion variables
        self.Z_movie_name_list = Z_movie_name_list
        self.Z_movie_input = Z_movie_input
        self.Z_imagedir = Z_imagedir
        self.start = start
        self.nb_prediction = nb_prediction
        self.fileextension = fileextension
        self.Z_n_tiles = Z_n_tiles
        self.overlap_percent = overlap_percent
        self.normalize = normalize

        self.model = self._build()

        # Z slice folder listener
        while 1:

            Z_Raw_path = os.path.join(self.Z_imagedir, self.fileextension)
            Z_filesRaw = glob.glob(Z_Raw_path)

            for Z_movie_name in Z_filesRaw:
                Z_Name = os.path.basename(os.path.splitext(Z_movie_name)[0])
                # Check for unique filename
                if Z_Name not in self.Z_movie_name_list:
                    self.Z_movie_name_list.append(Z_Name)
                    self.Z_movie_input.append(Z_movie_name)

                    if Z_Name in self.Z_movie_name_list:
                        self.Z_movie_name_list.remove(Z_Name)
                    if Z_movie_name in self.Z_movie_input:
                        self.Z_movie_input.remove(Z_movie_name)

            self.Z_movie_input_list = []
            for (k, v) in self.Z_movie_input.items():
                self.Z_movie_input_list.append(v)
            total_movies = len(self.Z_movie_input_list)

            if total_movies > self.start:
                current_movies = imread(
                    self.Z_movie_input_list[self.start : self.start + 1]
                )

                current_movies_down = current_movies
                # print(current_movies_down.shape)
                print(
                    "Predicting on Movie:",
                    self.Z_movie_input_list[self.start : self.start + 1],
                )

                eventboxes = []
                classedboxes = {}
                self.image = current_movies_down
                if self.normalize:
                    self.image = normalizeFloatZeroOne(self.image, 1, 99.8)
                # Break image into tiles if neccessary

                print("Doing ONEAT prediction")
                start_time = time.time()
                # Iterate over tiles

                for inputz in tqdm(range(0, self.image.shape[0])):
                    if inputz <= self.image.shape[0] - self.imagez:

                        eventboxes = []
                        classedboxes = {}
                        smallimage = CreateVolume(
                            self.image, self.imagez, inputz
                        )
                        predictions, allx, ally = self.predict_main(smallimage)
                        for p in range(0, len(predictions)):

                            sum_z_prediction = predictions[p]

                            if sum_z_prediction is not None:
                                # For each tile the prediction vector has shape N H W Categories + Training Vector labels
                                for i in range(0, sum_z_prediction.shape[0]):
                                    z_prediction = sum_z_prediction[i]
                                    boxprediction = focyoloprediction(
                                        ally[p],
                                        allx[p],
                                        z_prediction,
                                        self.stride,
                                        inputz,
                                        self.config,
                                        self.key_categories,
                                    )

                                    if boxprediction is not None:
                                        eventboxes = eventboxes + boxprediction

                        for (
                            event_name,
                            event_label,
                        ) in self.key_categories.items():

                            if event_label > 0:
                                current_event_box = []
                                for box in eventboxes:

                                    event_prob = box[event_name]
                                    if event_prob > 0:
                                        current_event_box.append(box)
                                classedboxes[event_name] = [current_event_box]

                        self.classedboxes = classedboxes
                        self.eventboxes = eventboxes

                        self.nms()
                        self.to_csv()
                        self.draw()

                print(
                    "____ Prediction took %s seconds ____ ",
                    (time.time() - start_time),
                )
                self.print_planes()
                self.genmap()
                self.start = self.start + 1
                self.predict(
                    self.imagedir,
                    self.Z_imagedir,
                    self.Z_movie_name_list,
                    self.Z_movie_input,
                    fileextension=self.fileextension,
                    nb_prediction=self.nb_prediction,
                    Z_n_tiles=self.Z_n_tiles,
                    overlap_percent=self.overlap_percent,
                )

    def _build(self):

        model_weights = os.path.join(self.model_dir, "weights.h5")
        Model = load_model(
            model_weights,
            custom_objects={"loss": self.yolo_loss, "Concat": Concat},
        )
        return Model

    def nms(self):

        best_iou_classedboxes = {}
        all_best_iou_classedboxes = {}
        self.all_iou_classedboxes = {}
        self.iou_classedboxes = {}
        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                # Get all events

                sorted_event_box = self.classedboxes[event_name][0]

                sorted_event_box = sorted(
                    sorted_event_box, key=lambda x: x[event_name], reverse=True
                )

                scores = [
                    sorted_event_box[i][event_name]
                    for i in range(len(sorted_event_box))
                ]
                best_sorted_event_box, all_boxes = simpleaveragenms(
                    sorted_event_box,
                    scores,
                    self.iou_threshold,
                    self.event_threshold,
                    event_name,
                )

                all_best_iou_classedboxes[event_name] = [all_boxes]
                best_iou_classedboxes[event_name] = [best_sorted_event_box]
        self.iou_classedboxes = best_iou_classedboxes
        self.all_iou_classedboxes = all_best_iou_classedboxes

    def genmap(self):

        image = imread(self.savename)
        Name = os.path.basename(os.path.splitext(self.savename)[0])
        Signal_first = image[:, :, :, 1]
        Signal_second = image[:, :, :, 2]
        Sum_signal_first = gaussian_filter(
            np.sum(Signal_first, axis=0), self.radius
        )
        Sum_signal_first = normalizeZeroOne(Sum_signal_first)
        Sum_signal_second = gaussian_filter(
            np.sum(Signal_second, axis=0), self.radius
        )

        Sum_signal_second = normalizeZeroOne(Sum_signal_second)

        Zmap = np.zeros(
            [Sum_signal_first.shape[0], Sum_signal_first.shape[1], 3]
        )
        Zmap[:, :, 0] = Sum_signal_first
        Zmap[:, :, 1] = Sum_signal_second
        Zmap[:, :, 2] = (Sum_signal_first + Sum_signal_second) / 2

        imwrite(self.basedirResults + Name + "_Zmap" + ".tif", Zmap)

    def to_csv(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:
                zlocations = []
                scores = []
                max_scores = []
                iou_current_event_box = self.iou_classedboxes[event_name][0]
                zcenter = iou_current_event_box["real_z_event"]
                max_score = iou_current_event_box["max_score"]
                score = iou_current_event_box[event_name]

                zlocations.append(zcenter)
                scores.append(score)
                max_scores.append(max_score)
                print(zlocations, scores)
                event_count = np.column_stack([zlocations, scores, max_scores])
                event_count = sorted(
                    event_count, key=lambda x: x[0], reverse=False
                )
                event_data = []
                csvname = (
                    self.basedirResults + "/" + event_name + "_FocusQuality"
                )
                writer = csv.writer(open(csvname + ".csv", "a"))
                filesize = os.stat(csvname + ".csv").st_size
                if filesize < 1:
                    writer.writerow(["Z", "Score", "Max_score"])
                for line in event_count:
                    if line not in event_data:
                        event_data.append(line)
                    writer.writerows(event_data)
                    event_data = []

    def fit_curve(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:
                readcsvname = (
                    self.basedirResults + "/" + event_name + "_FocusQuality"
                )
                self.dataset = pd.read_csv(readcsvname, delimiter=",")
                self.dataset_index = self.dataset.index

                Z = self.dataset[self.dataset.keys()[0]][1:]
                score = self.dataset[self.dataset.keys()[1]][1:]

                H, A, mu0, sigma = gauss_fit(np.array(Z), np.array(score))
                csvname = (
                    self.basedirResults
                    + "/"
                    + event_name
                    + "_GaussFitFocusQuality"
                )
                writer = csv.writer(open(csvname + ".csv", "a"))
                filesize = os.stat(csvname + ".csv").st_size
                if filesize < 1:
                    writer.writerow(["Amplitude", "Mean", "Sigma"])
                    writer.writerow([A, mu0, sigma])

                csvname = self.basedirResults + "/" + event_name

    def print_planes(self):
        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                csvfname = (
                    self.basedirResults
                    + "/"
                    + event_name
                    + "_focus_quality"
                    + ".csv"
                )
                dataset = pd.read_csv(csvfname, skiprows=0)
                z = dataset[dataset.keys()[0]][1:]
                score = dataset[dataset.keys()[1]][1:]
                maxz = z[np.argmax(score)] + 2
                print("Best Zs" + "for" + event_name + "at" + str(maxz))

    def draw(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:

                xlocations = []
                ylocations = []
                scores = []
                zlocations = []
                heights = []
                widths = []
                iou_current_event_boxes = self.all_iou_classedboxes[
                    event_name
                ][0]

                for iou_current_event_box in iou_current_event_boxes:

                    xcenter = iou_current_event_box["xcenter"]
                    ycenter = iou_current_event_box["ycenter"]
                    zcenter = iou_current_event_box["real_z_event"]
                    xstart = iou_current_event_box["xstart"]
                    ystart = iou_current_event_box["ystart"]
                    xend = xstart + iou_current_event_box["width"]
                    yend = ystart + iou_current_event_box["height"]
                    score = iou_current_event_box[event_name]

                    if event_label == 1:
                        for x in range(int(xstart), int(xend)):
                            for y in range(int(ystart), int(yend)):
                                if (
                                    y < self.image.shape[1]
                                    and x < self.image.shape[2]
                                ):
                                    self.Maskimage[int(zcenter), y, x, 1] = (
                                        self.Maskimage[int(zcenter), y, x, 1]
                                        + score
                                    )
                    else:

                        for x in range(int(xstart), int(xend)):
                            for y in range(int(ystart), int(yend)):
                                if (
                                    y < self.image.shape[1]
                                    and x < self.image.shape[2]
                                ):
                                    self.Maskimage[int(zcenter), y, x, 2] = (
                                        self.Maskimage[int(zcenter), y, x, 2]
                                        + score
                                    )

                    if score > 0.9:

                        xlocations.append(round(xcenter))
                        ylocations.append(round(ycenter))
                        scores.append(score)
                        zlocations.append(zcenter)
                        heights.append(iou_current_event_box["height"])
                        widths.append(iou_current_event_box["width"])

    def overlaptiles(self, sliceregion):

        if self.n_tiles == (1, 1):
            patch = []
            rowout = []
            column = []
            patchx = sliceregion.shape[2] // self.n_tiles[0]
            patchy = sliceregion.shape[1] // self.n_tiles[1]
            patchshape = (patchy, patchx)
            smallpatch, smallrowout, smallcolumn = chunk_list(
                sliceregion, patchshape, self.stride, [0, 0]
            )
            patch.append(smallpatch)
            rowout.append(smallrowout)
            column.append(smallcolumn)

        else:
            patchx = sliceregion.shape[2] // self.n_tiles[0]
            patchy = sliceregion.shape[1] // self.n_tiles[1]

            if patchx > self.imagex and patchy > self.imagey:
                if self.overlap_percent > 1 or self.overlap_percent < 0:
                    self.overlap_percent = 0.8

                jumpx = int(self.overlap_percent * patchx)
                jumpy = int(self.overlap_percent * patchy)

                patchshape = (patchy, patchx)
                rowstart = 0
                colstart = 0
                pairs = []
                # row is y, col is x

                while rowstart < sliceregion.shape[1] - patchy:
                    colstart = 0
                    while colstart < sliceregion.shape[2] - patchx:

                        # Start iterating over the tile with jumps = stride of the fully convolutional network.
                        pairs.append([rowstart, colstart])
                        colstart += jumpx
                    rowstart += jumpy

                # Include the last patch
                rowstart = sliceregion.shape[1] - patchy
                colstart = 0
                while colstart < sliceregion.shape[2]:
                    pairs.append([rowstart, colstart])
                    colstart += jumpx
                rowstart = 0
                colstart = sliceregion.shape[2] - patchx
                while rowstart < sliceregion.shape[1]:
                    pairs.append([rowstart, colstart])
                    rowstart += jumpy

                if (
                    sliceregion.shape[1] >= self.imagey
                    and sliceregion.shape[2] >= self.imagex
                ):

                    patch = []
                    rowout = []
                    column = []
                    for pair in pairs:
                        smallpatch, smallrowout, smallcolumn = chunk_list(
                            sliceregion, patchshape, self.stride, pair
                        )
                        if (
                            smallpatch.shape[1] >= self.imagey
                            and smallpatch.shape[2] >= self.imagex
                        ):
                            patch.append(smallpatch)
                            rowout.append(smallrowout)
                            column.append(smallcolumn)

            else:

                patch = []
                rowout = []
                column = []
                patchx = sliceregion.shape[2] // self.n_tiles[0]
                patchy = sliceregion.shape[1] // self.n_tiles[1]
                patchshape = (patchy, patchx)
                smallpatch, smallrowout, smallcolumn = chunk_list(
                    sliceregion, patchshape, self.stride, [0, 0]
                )
                patch.append(smallpatch)
                rowout.append(smallrowout)
                column.append(smallcolumn)
        self.patch = patch
        self.sy = rowout
        self.sx = column

    def predict_main(self, sliceregion):
        try:
            self.overlaptiles(sliceregion)
            predictions = []
            allx = []
            ally = []
            if len(self.patch) > 0:
                for i in range(0, len(self.patch)):

                    sum_time_prediction = self.make_patches(self.patch[i])
                    predictions.append(sum_time_prediction)
                    allx.append(self.sx[i])
                    ally.append(self.sy[i])

            else:

                sum_time_prediction = self.make_patches(self.patch)
                predictions.append(sum_time_prediction)
                allx.append(self.sx)
                ally.append(self.sy)

        except tf.errors.ResourceExhaustedError:

            print("Out of memory, increasing overlapping tiles for prediction")

            self.list_n_tiles = list(self.n_tiles)
            self.list_n_tiles[0] = self.n_tiles[0] + 1
            self.list_n_tiles[1] = self.n_tiles[1] + 1
            self.n_tiles = tuple(self.list_n_tiles)

            self.predict_main(sliceregion)

        return predictions, allx, ally

    def make_patches(self, sliceregion):

        predict_im = np.expand_dims(sliceregion, 0)

        prediction_vector = self.model.predict(
            np.expand_dims(predict_im, -1), verbose=0
        )

        return prediction_vector

    def make_batch_patches(self, sliceregion):

        prediction_vector = self.model.predict(
            np.expand_dims(sliceregion, -1), verbose=0
        )
        return prediction_vector


def chunk_list(image, patchshape, stride, pair):
    rowstart = pair[0]
    colstart = pair[1]

    endrow = rowstart + patchshape[0]
    endcol = colstart + patchshape[1]

    if endrow > image.shape[1]:
        endrow = image.shape[1]
    if endcol > image.shape[2]:
        endcol = image.shape[2]

    region = (
        slice(0, image.shape[0]),
        slice(rowstart, endrow),
        slice(colstart, endcol),
    )

    # The actual pixels in that region.
    patch = image[region]

    # Always normalize patch that goes into the netowrk for getting a prediction score

    return patch, rowstart, colstart


def CreateVolume(patch, imagez, timepoint):

    starttime = timepoint
    endtime = timepoint + imagez
    smallimg = patch[starttime:endtime, :]

    return smallimg


def normalizeZeroOne(x):
    x = x.astype("float32")

    minVal = np.min(x)
    maxVal = np.max(x)

    x = (x - minVal) / (maxVal - minVal + 1.0e-20)

    return x


def gauss(x, H, A, x0, sigma):
    return H + A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))


def gauss_fit(x, y):

    mean = sum(x * y) / sum(y)
    sigma = np.sqrt(sum(y * (x - mean) ** 2) / sum(y))
    popt, pcov = curve_fit(gauss, x, y, p0=[min(y), max(y), mean, sigma])
    return popt



import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
)
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QComboBox,
    QDoubleSpinBox,
    QFormLayout,
    QLabel,
    QPushButton,
    QSlider,
    QSpinBox,
    QWidget,
)


class OneatFrameWidget(QWidget):
    """Widget for interatviely making animations using the napari viewer."""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._layout = QFormLayout(parent=self)
        self.eventidbox = QComboBox()
        index = self.eventidbox.findText("linear", Qt.MatchFixedString)
        self.eventidbox.setCurrentIndex(index)

        self.plotidbox = QComboBox()
        index = self.plotidbox.findText("linear", Qt.MatchFixedString)
        self.plotidbox.setCurrentIndex(index)

        self.imageidbox = QComboBox()
        index = self.imageidbox.findText("linear", Qt.MatchFixedString)
        self.imageidbox.setCurrentIndex(index)

        self.heatstepsSpinBox = QSpinBox()
        self.heatstepsSpinBox.setValue(1)
        self.heatstepsSpinBox.setMaximum(100000)
        self.startprobSpinBox = QDoubleSpinBox()

        self.nmsspaceSpinBox = QDoubleSpinBox()
        self.nmsspaceSpinBox.setValue(10)
        self.nmsspaceSpinBox.setMaximum(100000)

        self.startprobSpinBox.setValue(0.9)
        self.startprobSpinBox.setDecimals(10)

        self.scoreSlider = QSlider(Qt.Horizontal, parent=self)
        self.scoreSlider.setToolTip("Scroll through probability score")
        self.scoreSlider.setRange(0, 5000)
        self.scoreSlider.setSingleStep(1)
        self.scoreSlider.setTickInterval(1)
        self.scoreSlider.setValue(0)

        self.label = QLabel(self)
        self.label.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)
        self.label.setMinimumWidth(80)
        self.label.setText(f"{0.9:.5f}")

        self.recomputeButton = QPushButton(
            "Recompute with changed parameters", parent=self
        )

        self.figure = plt.figure(figsize=(4, 4))
        self.multiplot_widget = FigureCanvas(self.figure)
        self.multiplot_widget.setMinimumSize(200, 200)
        self.ax = self.multiplot_widget.figure.subplots(1, 1)

        self._layout.addWidget(self.multiplot_widget)
        self._layout.addRow("Image/Movie", self.imageidbox)
        self._layout.addRow("Event", self.eventidbox)
        self._layout.addRow("Heat Map Steps", self.heatstepsSpinBox)
        self._layout.addRow("NMS space (px)", self.nmsspaceSpinBox)
        self._layout.addRow("Lowest probability event", self.startprobSpinBox)
        self._layout.addRow("Score slider", self.scoreSlider)
        self._layout.addRow("Score threshold", self.label)
        self._layout.addRow("Plot", self.plotidbox)
        self._layout.addRow(self.recomputeButton)



import argparse


class MidogConfig(argparse.Namespace):
    def __init__(self, cell_position_name, cell_position_label, **kwargs):

        self.cell_position_name = cell_position_name
        self.cell_position_label = cell_position_label
        assert len(cell_position_name) == len(cell_position_label)

    def to_json(self):

        configCord = {}

        for i in range(0, len(self.cell_position_name)):

            configCord[self.cell_position_name[i]] = self.cell_position_label[
                i
            ]

        return configCord
